from pathlib import Path, PurePath
import logging
from collections import namedtuple
from IPython.display import display, clear_output
import ipywidgets as widgets
from clease_gui import (utils, BaseDashboard, save_app_data, load_app_data,
                        update_statusbar, AppDataKeys)
from .logging_widget import (default_handler, set_all_levels, register_logger,
                             initialize_clease_gui_logging)

__all__ = ['display_ui', 'MainDashboard', 'get_assets_path']

logger = logging.getLogger(__name__)
register_logger(logger)


class TestAppData(BaseDashboard):
    def initialize(self):
        self.test_button = widgets.Button(description='See app data')
        self.test_button.on_click(self._on_click)
        self.output = widgets.Output()

    def display(self):
        display(self.test_button, self.output)

    def _on_click(self, b):
        from pprint import pprint
        data = self.app_data.copy()
        if 'settings' in data:
            data['settings'] = data['settings'].todict()
        with self.output:
            clear_output(wait=True)
            pprint(data)


class MainDashboard(BaseDashboard):
    def initialize(self):
        # We need to delay the import of these packages, to break cyclic imports
        from .status_bar import StatusBar
        from .db_viewer import DBViewerDashboard
        from .settings_maker import SettingsMakerDashboard
        from .new_structures import NewStructureDashboard
        from .calculate import CalculateDashboard
        from .fitting import FittingDashboard
        from .supercell import SupercellDashboard
        from .montecarlo import MainMCDashboard

        # The status bar should be the first thing to be initialized,
        # since it adds itself to the app_data
        self.status_bar = StatusBar(self.app_data)
        self.status_output = widgets.Output()
        with self.status_output:
            clear_output()
            self.status_bar.display()

        Tab = namedtuple('Tab', 'output,name')
        tabs = []  # List containing the tab objects to be added
        self.dashboards = {}  # Store the dashboards for internal access

        def add_tab(dashboard, display_name):
            """Helper function to add new tabs to the main dashboard"""
            output = widgets.Output()
            # Create the dashboard
            board = dashboard(self.app_data)
            name = display_name.lower().replace(' ', '_')
            self.dashboards[name] = board
            with output:
                clear_output()
                board.display()

            tab = Tab(output, display_name)
            tabs.append(tab)

        # Add main tabs
        add_tab(SettingsMakerDashboard, 'Settings')
        add_tab(NewStructureDashboard, 'New Structures')
        add_tab(DBViewerDashboard, 'Inspect DB')
        add_tab(CalculateDashboard, 'Calculate')
        add_tab(FittingDashboard, 'ECI')
        add_tab(SupercellDashboard, 'Supercell')
        add_tab(MainMCDashboard, 'Monte Carlo')
        if self.dev_mode:
            add_tab(TestAppData, 'Testing')

        # Add the tabs to the main dashboard
        self.tab = widgets.Tab(children=[tab.output for tab in tabs])

        # Add current working directory widet in the top
        cwd = Path('.').resolve()
        self.cwd_widget = widgets.Text(
            value=str(cwd),
            description='Current working directory:',
            layout=widgets.Layout(height="auto", width="100%"),
            style={'description_width': 'initial'},
        )
        # Update the app data, and observe any changes
        self.set_cwd(self.cwd_widget.value)
        self.cwd_widget.observe(self._on_cwd_change)
        self.logo = _get_logo_widget()

        for ii, tab in enumerate(tabs):
            self.tab.set_title(ii, tab.name)

        # Add button to clear logs
        clear = widgets.Button(description='Clear Logs')
        log_level = widgets.Dropdown(
            options=[
                ('Error', logging.ERROR),
                ('Warning', logging.WARNING),
                ('Info', logging.INFO),
                ('Debug', logging.DEBUG),
            ],
            value=logging.INFO,
            description='Logging level:',
            style={'description_width': 'initial'},
        )

        def on_clear_click(b):
            default_handler.clear_logs()

        clear.on_click(on_clear_click)

        def on_log_level_change(change):
            if utils.is_value_change(change):
                set_all_levels(change['new'])

        log_level.observe(on_log_level_change)
        initialize_clease_gui_logging()

        self.log_box = widgets.HBox(children=[clear, log_level])

        # save/load app state button
        self.save_app_state_button = widgets.Button(
            description='Save app data')
        self.save_app_state_button.on_click(self._on_save_app_data)
        self.load_app_state_button = widgets.Button(
            description='Load app data')
        self.load_app_state_button.on_click(self._on_load_app_data)

    def display(self):
        # Top bar box with logo and CWD widget
        top_bar_box = widgets.HBox(children=[self.logo, self.cwd_widget],
                                   layout={'align_items': 'center'})
        display(top_bar_box)
        display(self.tab)
        display(self.status_output)
        default_handler.show_logs()
        display(self.log_box)
        app_data_state_buttons = widgets.HBox(
            children=[self.save_app_state_button, self.load_app_state_button])
        display(app_data_state_buttons)

    @property
    def cwd(self):
        return self.app_data[self.KEYS.CWD]

    def set_cwd(self, value):
        self.app_data[self.KEYS.CWD] = Path(value).resolve()

    def _on_cwd_change(self, change) -> Path:
        if utils.is_value_change(change):
            self.set_cwd(change.new)

    @property
    def app_data_file_name(self):
        # Hardcode app data filename for now.
        return self.cwd / '.clease_gui_app_data.json'

    @update_statusbar
    def _on_save_app_data(self, b):
        try:
            fname = self.app_data_file_name
            logger.info('Saving app data to file: %s', str(fname))
            self._save_app_data(fname)
            logger.info('Save complete.')
        except Exception as exc:
            self.log_error(logger, exc)

    def _save_app_data(self, filename):
        save_app_data(self.app_data, filename)

    def _load_app_data(self, filename):
        new_data = load_app_data(filename)
        # Get the supercell, and make the supercell dashboard set it
        supercell = self.app_data.pop(self.KEYS.SUPERCELL, None)
        if supercell is not None:
            logger.debug('Updating supercell from app data')
            dashboard = self.dashboards['supercell']
            dashboard.set_supercell(supercell)

        self.app_data.update(new_data)
        # Try to update the settings page
        settings = new_data.get(self.KEYS.SETTINGS, None)
        if settings is not None:
            settings_dashboard = self.dashboards['settings']
            settings_dashboard.set_settings_widget_states(settings)

    @update_statusbar
    def _on_load_app_data(self, b):
        try:
            fname = self.app_data_file_name
            logger.info('Loading app data from file: %s', str(fname))
            self._load_app_data(fname)
            logger.info('Load complete.')
        except Exception as exc:
            self.log_error(logger, exc)


def display_ui(dev_mode=False):
    app_data = {AppDataKeys.DEV_MODE: dev_mode}
    main_dashboard = MainDashboard(app_data)
    main_dashboard.display()
    if dev_mode:
        # Return the app_data for inspection
        return app_data


def get_assets_path() -> PurePath:
    """Return the path to the assets directory"""
    return Path(__file__).parent / 'assets'


def _get_logo_widget():
    """Function to create the CLEASE logo and return it in an image widget"""
    logo_path = get_assets_path() / 'clease_logo.png'
    with logo_path.open('rb') as file:
        logo = file.read()
    return widgets.Image(
        value=logo,
        format='png',
    )



from IPython.display import display
import ipywidgets as widgets
from clease_gui import BaseDashboard
from . import CanonicalMC, PlotMCDashboard, MCRunDashboard

__all__ = ['MainMCDashboard']


class MainMCDashboard(BaseDashboard):
    def initialize(self):
        self.canonical_mc_out = widgets.Output()
        self.canonical_mc_dashboard = CanonicalMC(self.app_data)
        with self.canonical_mc_out:
            self.canonical_mc_dashboard.display()

        self.plot_out = widgets.Output(layout={'height': '100%'})
        self.plot_dashboard = PlotMCDashboard(self.app_data)
        with self.plot_out:
            self.plot_dashboard.display()

        self.store_mc_run_out = widgets.Output()
        self.store_mc_dashboard = MCRunDashboard(self.app_data)
        with self.store_mc_run_out:
            self.store_mc_dashboard.display()

        self.tab = widgets.Tab(children=[
            self.canonical_mc_out,
            self.plot_out,
            self.store_mc_run_out,
        ])

        self.tab.set_title(0, 'Canonical MC')
        self.tab.set_title(1, 'Plotting')
        self.tab.set_title(2, 'MC Runs')

    def display(self):
        display(self.tab)



import logging
from IPython.display import display
import ipywidgets as widgets
import clease
from clease_gui import BaseDashboard, register_logger, utils, update_statusbar
import clease_gui as gui

from . import fitting_widgets

logger = logging.getLogger(__name__)
register_logger(logger)

# eval_logger = logging.getLogger('clease.evaluate')
# register_logger(eval_logger)

__all__ = ['ECIFitDashboard']


class ECIFitDashboard(BaseDashboard):
    def initialize(self):

        self.fitting_output = widgets.Output()

        self.fitting_algo_dropdown = widgets.Dropdown(
            options=[
                ('LASSO',
                 fitting_widgets.LassoScheme(output=self.fitting_output)),
                (
                    'L2',
                    fitting_widgets.L2Scheme(output=self.fitting_output),
                ),
                # ('Genetic Algorithm', 'genetic_algorithm'),
            ],
            description='Fitting Algorithm:',
            **self.DEFAULT_STYLE_KWARGS,
        )
        self.fitting_algo_dropdown.value.display()

        def _on_fitting_change(change):
            if utils.is_value_change(change):
                change.new.display()

        self.fitting_algo_dropdown.observe(_on_fitting_change)

        self.scoring_output = widgets.Output()
        self.scoring_widget = fitting_widgets.ScoringSchemeWidget()

        self.db_select_out = widgets.Output()
        self.db_select_widget = fitting_widgets.DBSelectWidget()

        self.fit_eci_button = widgets.Button(description='Fit ECI',
                                             button_style='primary')
        self.fit_eci_button.on_click(self._on_fit_eci_click)

    def display(self):
        # self.update_algo_widgets()
        self.scoring_widget.display_widgets(output=self.scoring_output)
        self.db_select_widget.display_widgets(output=self.db_select_out)
        display(
            self.fitting_algo_dropdown,
            self.fitting_output,
            self.scoring_output,
            self.db_select_out,
            self.fit_eci_button,
        )

    @property
    def fitting_algo_widget(self) -> fitting_widgets.RegressionScheme:
        return self.fitting_algo_dropdown.value

    def get_db_name(self):
        db_name = self.db_select_widget.get_db_name()
        if db_name is not None:
            cwd = self.app_data[self.KEYS.CWD]
            return cwd / db_name
        # No db name was specified, try getting it from settings
        return self.get_db_name()

    def _make_evaluator(self):
        settings = self.settings
        scoring_scheme_kwargs = self.scoring_widget.value

        evaluator = clease.Evaluate(
            settings,
            **scoring_scheme_kwargs,
        )
        return evaluator

    def fit_eci(self):
        # Get all parameters
        wdgt = self.fitting_algo_widget
        keys = list(wdgt.HYPERPARAMS)
        if keys != ['alpha']:
            raise NotImplementedError('Not yet implemented yet. Got keys:',
                                      keys)

        alpha = wdgt.hyper_widget_dct['alpha']
        is_opt = alpha.is_optimization

        evaluator = self._make_evaluator()
        scheme = wdgt.get_fitting_scheme()
        evaluator.set_fitting_scheme(fitting_scheme=scheme)
        if is_opt:
            logger.info('Optimizing alpha')
            min_val, max_val, step_val = alpha.get_min_max_step()
            spacing_type = alpha.get_spacing_type()

            scale = 'log' if spacing_type == 'logspace' else 'linear'
            logger.debug('Scale: %s', scale)
            alphas, cv = evaluator.alpha_CV(
                alpha_min=min_val,
                alpha_max=max_val,
                num_alpha=step_val,
                scale=scale,
            )
            # Get optimial alpha
            ind = cv.argmin()
            alpha_val = alphas[ind]
            cv_min = cv[ind]
            logger.info('Optimized alpha: %.3e, cv score: %.3e', alpha_val,
                        cv_min)
        else:
            alpha_val = alpha.get_float_value()
        logger.info('Running fit with alpha: %.3e', alpha_val)
        # Even if we did an optimization, we want to fit at the optimized
        evaluator.set_fitting_scheme(fitting_scheme=scheme, alpha=alpha_val)
        evaluator.get_eci()
        self.set_evaluator(evaluator)
        eci = evaluator.get_eci_dict()
        self.set_eci(eci)
        logger.info('Fitting complete.')

    def set_evaluator(self, evaluator):
        """The evaluator cannot be saved to file, so save it as private"""

        # Make a shallow copy, so that it doesn't suddenly change.
        self.app_data[self.KEYS.EVALUATE] = evaluator

    @update_statusbar
    @gui.utils.disable_cls_widget('fit_eci_button')
    def _on_fit_eci_click(self, b):
        try:
            self.fit_eci()
        except Exception as exc:
            self.log_error(logger, 'Failed to fit: %s', exc)

    def set_eci(self, value):
        self.app_data[self.KEYS.ECI] = value

    @property
    def eci(self):
        try:
            return self.app_data[self.KEYS.ECI]
        except KeyError:
            raise ValueError(
                'No ECI\'s present. Run fit first, or load ECI from file.')



from typing import Dict, Any
from abc import ABC, abstractmethod
from IPython.display import display
import ipywidgets as widgets
import ase
import clease
from clease_gui import WidgetStyleMixin, utils
import clease_gui as gui


class SimpleWidget(WidgetStyleMixin, ABC):
    def __init__(self, app_data):
        self.app_data = app_data
        self._widget_dct = self._make_widgets()

    @property
    def widget_dct(self):
        return self._widget_dct

    def get_values(self) -> Dict[str, Any]:
        return {key: self.get_value(key) for key in self.widget_dct.keys()}

    def get_value(self, key: str) -> Any:
        return self.widget_dct[key].value

    def display(self):
        display(*self.widget_dct.values())

    @abstractmethod
    def _make_widgets(self):
        """Construct the widget(s)"""


class AtomsWidget(SimpleWidget):
    """Widget for passing in an atoms object stored on disk."""
    def _make_widgets(self):
        self.atoms_widget = widgets.Text(
            description='Atoms (filename):',
            value='',
            **self.DEFAULT_STYLE_KWARGS,
        )

        label = widgets.Label(
            'Template for structure generation. Leave empty to use a random template.'
        )
        box = widgets.HBox(children=[self.atoms_widget, label])

        return {'atoms': box}

    def get_value(self, key):
        if key == 'atoms':
            value = self.atoms_widget.value
            if value == '':
                return None
            cwd = self.app_data[gui.AppDataKeys.CWD]
            return ase.io.read(cwd / value, index=-1)
        return super().get_value(key)


class GSAtomsWidget(SimpleWidget):
    """Atoms widget for ground structure. Slightly different from the regular
    atoms widget """
    def _make_widgets(self):
        self.atoms_widget = widgets.Text(
            description='Atoms (filename):',
            value='',
            **self.DEFAULT_STYLE_KWARGS,
        )

        label = widgets.Label(
            'Template for structure generation. Leave empty to use a random template.'
        )
        box = widgets.HBox(children=[self.atoms_widget, label])

        return {'atoms': box}

    @property
    def settings(self) -> clease.settings.ClusterExpansionSettings:
        return self.app_data[gui.AppDataKeys.SETTINGS]

    def get_value(self, key):
        if key == 'atoms':
            value = self.atoms_widget.value
            if value == '':
                # TODO: Do something better than a random template
                atoms = self.settings.template_atoms.weighted_random_template()
                return atoms
            cwd = self.app_data[gui.AppDataKeys.CWD]
            return ase.io.read(cwd / value, index=-1)
        return super().get_value(key)


class ProbeStructTemp(SimpleWidget):
    """Widget specifically for temperature estimation for ProbeStructure, since it
    can have init_temp and final_temp=None, which turns on automatic temperature
    estimation. This requires having the init_temp and final_temp widgets linked to
    a checkbox"""
    def _make_widgets(self):

        self.auto_temp = widgets.Checkbox(
            description='Automatically determine temperature?',
            value=True,
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.init_temp = widgets.BoundedFloatText(
            value=2000.0,
            description='Initial temperature:',
            min=0,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        )

        self.final_temp = widgets.BoundedFloatText(
            value=1.0,
            description='Final temperature:',
            min=0,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        )

        def _on_change_auto(change):
            if not utils.is_value_change(change):
                return
            self._update_widgets()

        self._update_widgets()
        self.auto_temp.observe(_on_change_auto)

        return dict(init_temp=self.init_temp, final_temp=self.final_temp)

    def display(self):
        display(self.auto_temp)
        super().display()

    def get_value(self, key):
        auto = self.auto_temp.value
        if key == 'init_temp' or key == 'final_temp':
            auto = self.auto_temp.value
            if auto:
                return None
        return super().get_value(key)

    def _update_widgets(self):
        """Update the widgets according to whether
        we automatically determine temp"""
        # If auto = True, we disable
        disabled = self.auto_temp.value

        self.init_temp.disabled = disabled
        self.final_temp.disabled = disabled


class InitTemp(SimpleWidget):
    def _make_widgets(self):
        return dict(init_temp=widgets.BoundedFloatText(
            value=2000.0,
            description='Initial temperature:',
            min=0,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        ))


class FinalTemp(SimpleWidget):
    def _make_widgets(self):
        return dict(final_temp=widgets.BoundedFloatText(
            value=1.0,
            description='Final temperature:',
            min=0,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        ))


class NumTemp(SimpleWidget):
    def _make_widgets(self):
        return dict(num_temp=widgets.BoundedIntText(
            value=10,
            description='Num. temperatures:',
            min=1,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        ))


class NumStepsPerTemp(SimpleWidget):
    def _make_widgets(self):
        return dict(num_steps_per_temp=widgets.BoundedIntText(
            value=1000,
            description='Num. steps per temp:',
            min=1,
            max=1e8,
            **self.DEFAULT_STYLE_KWARGS,
        ))


class ApproxMeanVar(SimpleWidget):
    def _make_widgets(self):
        dct = {}
        default_value = True
        mean_var = widgets.Checkbox(
            value=default_value,
            description='Approximate mean variance?',
            **self.DEFAULT_STYLE_KWARGS,
        )
        dct['approx_mean_var'] = mean_var

        num_samples_var = widgets.IntText(
            value=10000,
            disabled=not default_value,
            description='Num. samples for approx.',
            **self.DEFAULT_STYLE_KWARGS,
        )
        dct['num_samples_var'] = num_samples_var

        # num_samples_var is only enabled for approx_mean_var == True
        def on_mean_var_change(change):
            if utils.is_value_change(change):
                num_samples_var.disabled = change['new'] is not True

        mean_var.observe(on_mean_var_change)

        return dct


class RandomComposition(SimpleWidget):
    def _make_widgets(self):
        return dict(random_composition=widgets.Checkbox(
            value=True,
            description='Random composition?',
            **self.DEFAULT_STYLE_KWARGS,
        ))



from setuptools import setup, find_packages
from pathlib import Path


def read_requirements(filename):
    with Path(__file__).with_name(filename).open('r') as f:
        return tuple(line.strip() for line in f if line.strip())


publish_requirements = read_requirements('publish_requirements.txt')
test_requirements = read_requirements('test_requirements.txt')
doc_requirements = read_requirements('doc_requirements.txt')
dev_requirements = read_requirements('dev_requirements.txt')

EXTRAS_REQUIRE = {
    'dev': dev_requirements,
    'publish': publish_requirements,
    'test': test_requirements,
    'doc': doc_requirements,
}

EXTRAS_REQUIRE['all'] = set(value for tup in EXTRAS_REQUIRE.values()
                            for value in tup)

setup(
    packages=find_packages(include=['clease_gui', 'clease_gui.*']),
    extras_require=EXTRAS_REQUIRE,
)
